from oem.media.show.identifier import EpisodeIdentifier


class EpisodeMatch(EpisodeIdentifier):
    def __init__(self, identifiers, season_num=None, episode_num=None, absolute_num=None,
                 progress=None, part=None, mappings=None):

        super(EpisodeMatch, self).__init__(
            season_num=season_num,
            episode_num=episode_num,
            absolute_num=absolute_num,

            progress=progress,
            part=part
        )

        self.identifiers = identifiers or {}
        self.mappings = mappings or []

    @property
    def valid(self):
        return len(self.identifiers) > 0 and (
            super(EpisodeMatch, self).valid or
            len(self.mappings) > 0
        )

    def to_dict(self):
        result = super(EpisodeMatch, self).to_dict()

        result['identifiers'] = self.identifiers

        if self.mappings:
            result['mappings'] = [m.to_dict(compact=False) for m in self.mappings]

        return result

    def __repr__(self):
        fragments = []

        # Identifiers
        if self.identifiers:
            fragments.append(
                '(' + (', '.join(
                    ('%s: %r' % (key, value)) for key, value in self.identifiers.items()
                )) + ')'
            )

            if self.absolute_num is not None or self.season_num is not None:
                fragments.append('-')

        # Absolute
        if self.absolute_num is not None:
            fragments.append('[%03d]' % self.absolute_num)

        # Season + Episode
        if self.season_num is not None and self.episode_num is not None:
            fragments.append('S%02dE%02d' % (self.season_num, self.episode_num))
        elif self.season_num is not None:
            fragments.append('S%02d' % self.season_num)

        # Attributes
        attributes = [
            ('%s: %r' % (key, getattr(self, key))) for key in ['progress'] if getattr(self, key)
        ]

        if attributes:
            fragments.append('(%s)' % (', '.join(attributes)))

        return '<%s%s>' % (
            self.__class__.__name__,
            (' %s' % (' '.join(fragments))) if fragments else ''
        )



from oem_core.core.plugin import PluginManager
from oem_framework.plugin import Plugin
from oem_framework.storage import ProviderStorage


class Provider(Plugin):
    def __init__(self, storage):
        self._storage = storage

        self._client = None

    #
    # Properties
    #

    @property
    def client(self):
        return self._client

    @property
    def formats(self):
        return self._client.formats

    @property
    def plugins(self):
        return self._client.plugins

    @property
    def storage(self):
        return self._storage

    #
    # Public methods
    #

    def initialize(self, client):
        self._client = client

        self._storage = self._construct_storage(self._storage)

    #
    # Abstract methods
    #

    def fetch(self, source, target, key, metadata):
        raise NotImplementedError

    def open_database(self, source, target):
        raise NotImplementedError

    #
    # Private methods
    #

    def _construct_storage(self, storage_or_key):
        if isinstance(storage_or_key, ProviderStorage):
            # Use provided source
            storage = storage_or_key
        elif PluginManager.has('storage', storage_or_key):
            # Construct source by key
            storage = PluginManager.get('storage', storage_or_key)()
        else:
            raise ValueError('Unknown storage interface: %r' % storage_or_key)

        # Initialize source
        storage.initialize(self._client)
        return storage



# from oem.media.movie import MovieMatch
# from oem.media.show import EpisodeMatch
#
#
# def pytest_assertrepr_compare(op, left, right):
#     if op == "==" and (isinstance(left, EpisodeMatch) or isinstance(right, EpisodeMatch)):
#         return [
#             '%r == %r' % (left, right)
#         ]
#
#     if op == "==" and (isinstance(left, MovieMatch) or isinstance(right, MovieMatch)):
#         return [
#             '%r == %r' % (left, right)
#         ]



from oem_framework.core.elapsed import Elapsed
from oem_framework.plugin import Plugin

import logging

log = logging.getLogger(__name__)


class Service(Plugin):
    __databases__ = {}
    __packages__ = {}
    __services__ = {}

    def __init__(self, client, source, target, formats=None):
        self._client = client
        self._source = source
        self._target = target
        self._formats = formats

        self._database = None
        self._collection = None
        self._loaded = False

    @property
    def database_name(self):
        return 'oem_database_%s_%s' % (self._source, self._target)

    @property
    def loaded(self):
        return self._loaded

    @property
    def package_name(self):
        return 'oem-database-%s-%s' % (self._source, self._target)

    @property
    def provider(self):
        return self._client._provider

    @property
    def source_key(self):
        return self._source

    @property
    def target_key(self):
        return self._target

    @Elapsed.track
    def load(self):
        if self._loaded:
            return True

        # Load database
        self._database = self.provider.open_database(
            self.source_key,
            self.target_key
        )

        if self._database is None:
            log.warn('Unable to load database for: %s -> %s', self.source_key, self.target_key)
            return False

        # Load collection
        self._collection = self._database.load_collection(
            self._source,
            self._target
        )

        if self._collection is None:
            log.warn('Unable to load collection for: %s -> %s', self.source_key, self.target_key)
            return False

        # Successfully loaded service
        log.info('Loaded service: %-5s -> %-5s (storage: %r)', self._source, self._target, self._database.storage)
        self._loaded = True

        return True

    @Elapsed.track
    def fetch(self, key, metadata):
        # Ensure database is loaded
        if not self.load():
            return False

        # Ensure item is loaded
        return self.provider.fetch(
            self.source_key,
            self.target_key,
            key, metadata
        )



from oem.core.exceptions import AbsoluteNumberRequiredError  # NOQA

import logging

log = logging.getLogger(__name__)


try:
    from oem.client import Client
    OemClient = Client
except ImportError as ex:
    log.warn('Unable to import client - %s', ex, exc_info=True)
    Client = None
    OemClient = None



class Identifier(object):
    @property
    def valid(self):
        return False

    def to_dict(self):
        raise NotImplementedError

    def to_frozenset(self, data=None):
        if data is None:
            data = self.to_dict()

        if type(data) is dict:
            data = data.items()

        result = []

        for item in data:
            if type(item) is tuple and len(item) == 2:
                key, value = item
            else:
                key = None
                value = item

            if type(value) is dict:
                value = self.to_frozenset(value)

            if type(value) is list:
                value = self.to_frozenset(value)

            if key is not None:
                result.append((key, value))
            else:
                result.append(value)

        return frozenset(result)

    def __hash__(self):
        return hash(self.to_frozenset())

    def __eq__(self, other):
        if not other:
            return False

        return self.__hash__() == other.__hash__()

    def __ne__(self, other):
        return not(self == other)



from oem.media.movie.identifier import MovieIdentifier
from oem.media.movie.match import MovieMatch
from oem.media.show import EpisodeIdentifier

from copy import deepcopy


class MovieMapper(object):
    def __init__(self, service):
        self._service = service

    def match(self, movie, identifier, resolve_mappings=True):
        if identifier is not None and not identifier.valid:
            raise ValueError('Invalid value provided for "identifier" parameter')

        if isinstance(identifier, EpisodeIdentifier) and identifier.season_num != 1:
            return None

        # Movie
        best = self._match_movie(movie, identifier)

        # Part
        result = self._match_part(movie, identifier)

        if result:
            best = result

        # Return best result
        return best

    def _match_movie(self, movie, identifier):
        if isinstance(identifier, MovieIdentifier) and identifier.part is not None and identifier.part > 1:
            return None

        if isinstance(identifier, EpisodeIdentifier) and identifier.episode_num != 1:
            return None

        # Retrieve progress
        progress = None

        if identifier:
            progress = identifier.progress

        # Create movie match
        return MovieMatch(
            self._get_identifiers(movie),
            progress=progress
        )

    def _match_part(self, movie, identifier):
        if isinstance(identifier, MovieIdentifier):
            part_num = identifier.part
        elif isinstance(identifier, EpisodeIdentifier):
            part_num = identifier.episode_num
        else:
            part_num = 1

        # Retrieve part
        part = movie.parts.get(str(part_num))

        if not part:
            return None

        # Retrieve progress
        progress = None

        if identifier:
            progress = identifier.progress

        # Create movie match
        return MovieMatch(
            self._get_identifiers(part),
            progress=progress
        )

    def _get_identifiers(self, movie):
        # Retrieve identifiers from objects
        identifiers = movie.identifiers

        # Copy identifiers
        if identifiers:
            identifiers = deepcopy(identifiers)
        else:
            identifiers = {}

        # Remove source identifier
        if self._service.source_key in identifiers:
            del identifiers[self._service.source_key]

        return identifiers



from oem.media.core.base.identifier import Identifier


class EpisodeIdentifier(Identifier):
    __slots__ = ['season_num', 'episode_num', 'absolute_num', 'progress']

    def __init__(self, season_num=None, episode_num=None, absolute_num=None, progress=None, part=None):
        # Season + Episode Identifier
        self.season_num = season_num
        self.episode_num = episode_num

        # Absolute Identifier
        self.absolute_num = absolute_num

        # Extra
        self.progress = progress
        self.part = part

    @property
    def valid(self):
        return (
            self.season_num is not None and
            self.episode_num is not None
        ) or (
            self.absolute_num is not None
        )

    def to_dict(self):
        result = {}

        if self.absolute_num is not None:
            result['absolute_num'] = self.absolute_num

        if self.season_num is not None:
            result['season_num'] = self.season_num

        if self.episode_num is not None:
            result['episode_num'] = self.episode_num

        if self.progress is not None:
            result['progress'] = self.progress

        if self.part is not None:
            result['part'] = self.part

        return result

    def __repr__(self):
        attributes = [
            ('%s: %r' % (key, getattr(self, key)))
            for key in ['progress', 'part'] if getattr(self, key)
        ]

        fragments = []

        if self.absolute_num:
            fragments.append('[%03d]' % self.absolute_num)

        if self.season_num is not None and self.episode_num is not None:
            fragments.append('S%02dE%02d' % (self.season_num, self.episode_num))
        elif self.season_num is not None:
            fragments.append('S%02d' % self.season_num)

        if attributes:
            fragments.append('(%s)' % (', '.join(attributes)))

        return '<%s%s>' % (
            self.__class__.__name__,
            (' %s' % (' '.join(fragments))) if fragments else ''
        )



from oem.media.movie.identifier import MovieIdentifier


class MovieMatch(MovieIdentifier):
    def __init__(self, identifiers, progress=None):
        super(MovieMatch, self).__init__(progress)

        self.identifiers = identifiers or {}

    def to_dict(self):
        result = super(MovieMatch, self).to_dict()

        result['identifiers'] = self.identifiers

        return result

    def __repr__(self):
        fragments = []

        # Identifiers
        if self.identifiers:
            fragments.append(
                '(' + (', '.join(
                    ('%s: %r' % (key, value))
                    for key, value in self.identifiers.items()
                )) + ')'
            )

        # Attributes
        attributes = [
            ('%s: %r' % (key, getattr(self, key))) for key in ['progress'] if getattr(self, key)
        ]

        if attributes:
            fragments.append('(%s)' % (', '.join(attributes)))

        return '<%s%s>' % (
            self.__class__.__name__,
            (' %s' % (' '.join(fragments))) if fragments else ''
        )



from oem.core.exceptions import AbsoluteNumberRequiredError
from oem.media.movie.identifier import MovieIdentifier
from oem.media.show.identifier import EpisodeIdentifier
from oem.media.show.match import EpisodeMatch
from oem_framework.core.helpers import try_convert

from copy import deepcopy
import logging

log = logging.getLogger(__name__)


class ShowMapper(object):
    def __init__(self, service):
        self._service = service

    def match(self, show, identifier, resolve_mappings=True):
        if identifier is None:
            # Create identifier for S01E01
            identifier = EpisodeIdentifier(1, 1)
        elif isinstance(identifier, MovieIdentifier):
            # Convert movie identifier to S01E01
            identifier = EpisodeIdentifier(1, 1, progress=identifier.progress)

        # Validate identifier
        if identifier and not identifier.valid:
            raise ValueError('Invalid value provided for "identifier" parameter')

        # Show
        best = self._match_show(show, identifier)

        # Season
        season, result = self._match_season(show, identifier)

        if result:
            best = result

        if season:
            # Episode
            result = self._match_episode(
                show, season, identifier,
                resolve_mappings=resolve_mappings
            )

            if result:
                best = result

        # Return best result
        return best

    def _match_show(self, show, identifier):
        # Retrieve "default_season" parameter
        default_season = None

        if 'default_season' in show.parameters:
            default_season = show.parameters['default_season']

            if default_season != 'a':
                # Cast season number to an integer
                default_season = try_convert(default_season, int)

                if default_season is None:
                    log.warn(
                        'Invalid value provided for the "default_season" parameter: %r',
                        show.parameters['default_season']
                    )
                    return None

        # Retrieve season number
        season_num = identifier.season_num

        if season_num is None or default_season is None or default_season == 'a':
            season_num = default_season
        elif season_num > 0:
            season_num = default_season + (season_num - 1)

        # Retrieve episode number
        episode_num = identifier.episode_num

        if 'episode_offset' in show.parameters:
            episode_num += int(show.parameters['episode_offset'])

        # Build episode match
        if season_num != 'a':
            match = EpisodeMatch(
                self._get_identifiers(show),
                season_num=season_num,
                episode_num=episode_num,

                progress=identifier.progress
            )
        else:
            if identifier.absolute_num is None:
                raise AbsoluteNumberRequiredError('Unable to match %r, an absolute number is required' % identifier)

            match = EpisodeMatch(
                self._get_identifiers(show),
                absolute_num=identifier.absolute_num,

                progress=identifier.progress
            )

        if not match.valid:
            return None

        return match

    def _match_season(self, show, identifier):
        # Try retrieve matching season
        season = show.seasons.get(str(identifier.season_num)) or show.seasons.get('a')

        if not season:
            return None, None

        if season.number == 'a':
            if identifier.absolute_num is None:
                raise AbsoluteNumberRequiredError('Unable to match %r, an absolute number is required' % identifier)

            return season, EpisodeMatch(
                self._get_identifiers(show, season),
                absolute_num=identifier.absolute_num,

                progress=identifier.progress
            )

        # Look for matching season mapping
        for season_mapping in season.mappings:
            if not (season_mapping.start <= identifier.episode_num <= season_mapping.end):
                continue

            return season, EpisodeMatch(
                self._get_identifiers(show, season),
                int(season_mapping.season),
                identifier.episode_num + season_mapping.offset,

                progress=identifier.progress
            )

        # Retrieve "default_season" parameter
        default_season = None

        if 'default_season' in season.parameters:
            default_season = season.parameters['default_season']

            if default_season != 'a':
                # Cast season number to an integer
                default_season = try_convert(default_season, int)

                if default_season is None:
                    log.warn(
                        'Invalid value provided for the "default_season" parameter: %r',
                        season.parameters['default_season']
                    )
                    return season, None

        # Retrieve season number
        season_num = identifier.season_num

        if season.identifiers:
            season_num = 1

        if default_season is not None:
            season_num = default_season

        # Retrieve episode number
        episode_num = identifier.episode_num

        # Apply episode offset
        episode_offset = self._get_parameter('episode_offset', show, season)

        if episode_offset is not None:
            episode_num += int(episode_offset)

        # Build season match
        match = EpisodeMatch(
            self._get_identifiers(show, season),
            season_num=season_num,
            episode_num=episode_num,

            progress=identifier.progress
        )

        if not match.valid:
            return season, None

        return season, match

    def _match_episode(self, show, season, identifier, resolve_mappings=True):
        episode = season.episodes.get(str(identifier.episode_num))

        if not episode:
            return None

        if not resolve_mappings:
            match = EpisodeMatch(
                self._get_identifiers(show, season, episode),
                mappings=episode.mappings
            )

            if not match.valid:
                return None

            return match

        if identifier.part is not None and identifier.part - 1 < len(episode.mappings):
            # Parse episode mapping
            valid, match = self._parse_episode_mapping(
                show, season, episode, episode.mappings[identifier.part - 1],
                part=identifier.part
            )

            if valid:
                return match

        for episode_mapping in episode.mappings:
            # Parse timeline attributes
            progress = identifier.progress

            if episode_mapping.timeline:
                if identifier.progress is None:
                    raise ValueError('Missing required parameter "progress"')

                if 'source' in episode_mapping.timeline:
                    timeline_source = episode_mapping.timeline['source']

                    if not (timeline_source.start <= identifier.progress <= timeline_source.end):
                        continue

                    # Calculate progress
                    progress = (
                        float(identifier.progress - timeline_source.start) *
                        (100 / (timeline_source.end - timeline_source.start))
                    )
                elif 'target' in episode_mapping.timeline:
                    timeline_target = episode_mapping.timeline['target']

                    # Calculate progress
                    progress = (
                        timeline_target.start + (
                            float(identifier.progress) /
                            (100 / (timeline_target.end - timeline_target.start))
                        )
                    )

            # Parse episode mapping
            valid, match = self._parse_episode_mapping(
                show, season, episode, episode_mapping,
                progress=progress
            )

            if valid:
                return match

        return None

    def _parse_episode_mapping(self, show, season, episode, episode_mapping, progress=None, part=None):
        # Parse mapping attributes
        try:
            season_num = int(episode_mapping.season)
        except:
            return False, None

        try:
            episode_num = int(episode_mapping.number)
        except:
            return False, None

        # Return episode match
        match = EpisodeMatch(
            self._get_identifiers(show, season, episode),
            season_num=season_num,
            episode_num=episode_num,

            progress=progress,
            part=part
        )

        if not match.valid:
            return True, None

        return True, match

    def _get_identifiers(self, show, season=None, episode=None):
        # Retrieve identifiers from objects
        if show and season and episode:
            identifiers = episode.identifiers or season.identifiers or show.identifiers
        elif show and season:
            identifiers = season.identifiers or show.identifiers
        else:
            identifiers = show.identifiers

        # Copy identifiers
        if identifiers:
            identifiers = deepcopy(identifiers)
        else:
            identifiers = {}

        # Remove source identifier
        if self._service.source_key in identifiers:
            del identifiers[self._service.source_key]

        return identifiers

    def _get_parameter(self, key, show, season=None, episode=None):
        for obj in [episode, season, show]:
            if not obj:
                continue

            return obj.parameters.get(key)

        return None



from tests.core.mock import MockService

from oem.core.exceptions import AbsoluteNumberRequiredError
from oem.media.show import ShowMapper, EpisodeIdentifier
from oem_core.models import Collection, Show, Season, Episode, EpisodeMapping

import pytest

collection = Collection(None, 'anidb', 'tvdb')
service = MockService('anidb', 'tvdb')


#
# Show
#


def test_show_invalid_default_season():
    mapper = ShowMapper(service)

    assert mapper.match(
        Show(collection, {}, [], default_season='INVALID'),
        EpisodeIdentifier(1, 1)
    ) is None


#
# Season
#


def test_season_absolute_number_required():
    mapper = ShowMapper(service)

    show = Show(collection, {}, [])
    show.seasons['a'] = Season(collection, show, 'a')

    with pytest.raises(AbsoluteNumberRequiredError):
        mapper.match(show, EpisodeIdentifier(1, 1))


def test_season_invalid_default_season():
    mapper = ShowMapper(service)

    show = Show(collection, {}, [])
    show.seasons['1'] = Season(collection, show, '1', default_season='INVALID')

    assert mapper.match(show, EpisodeIdentifier(1, 1)) is None


def test_season_invalid_match():
    mapper = ShowMapper(service)

    show = Show(collection, {}, [])
    show.seasons['1'] = Season(collection, show, '1')

    assert mapper.match(show, EpisodeIdentifier(1, 1)) is None


#
# Episode
#


def test_episode_invalid_mapping_season():
    mapper = ShowMapper(service)

    show = Show(collection, {}, [])
    show.seasons['1'] = Season(collection, show, '1')
    show.seasons['1'].episodes['1'] = Episode(collection, show.seasons['1'], '1')
    show.seasons['1'].episodes['1'].mappings = [
        EpisodeMapping(collection, show.seasons['1'].episodes['1'], 'INVALID', '1')
    ]

    assert mapper.match(show, EpisodeIdentifier(1, 1)) is None


def test_episode_invalid_mapping_episode():
    mapper = ShowMapper(service)

    show = Show(collection, {}, [])
    show.seasons['1'] = Season(collection, show, '1')
    show.seasons['1'].episodes['1'] = Episode(collection, show.seasons['1'], '1')
    show.seasons['1'].episodes['1'].mappings = [
        EpisodeMapping(collection, show.seasons['1'].episodes['1'], '1', 'INVALID')
    ]

    assert mapper.match(show, EpisodeIdentifier(1, 1)) is None


def test_episode_invalid_match():
    mapper = ShowMapper(service)

    show = Show(collection, {}, [])
    show.seasons['1'] = Season(collection, show, '1')
    show.seasons['1'].episodes['1'] = Episode(collection, show.seasons['1'], '1')
    show.seasons['1'].episodes['1'].mappings = [
        EpisodeMapping(collection, show.seasons['1'].episodes['1'], '1', '1')
    ]

    assert mapper.match(show, EpisodeIdentifier(1, 1)) is None



class MockService(object):
    def __init__(self, source, target):
        self._source = source
        self._target = target

    @property
    def source_key(self):
        return self._source

    @property
    def target_key(self):
        return self._target



from oem.core.providers.base import Provider
from oem.version import __version__
from oem_core.core.plugin import PluginManager

import inspect
import logging
import six

log = logging.getLogger(__name__)


class Client(object):
    version = __version__

    def __init__(self, services, provider, formats=None):
        """OpenEntityMap (OEM) Client

        :param services: List of services to load (e.g. "anidb")
        :type services: list

        :param provider: Provider to use for databases (e.g. "package", "release/incremental")
        :type provider: str or oem.core.providers.base.Base

        :param formats: List of formats to use, or `None` for any
        :type formats: list or None
        """

        self._formats = formats

        # Discover available plugins
        self._plugins = PluginManager
        self._plugins.discover()

        # Construct plugins
        self._services = self._construct_services(services)
        self._provider = self._construct_provider(provider)

        # Build database + package tables
        self._databases = {}
        self._packages = {}

        for _, cls in self._load_plugins('client', services, construct=False):
            # Merge service databases into client
            if cls.__databases__:
                self._databases.update(cls.__databases__)
            else:
                log.warn('Service %r has no "__databases__" defined', cls.__key__)

            # Merge service packages into client
            if cls.__packages__:
                self._packages.update(cls.__packages__)
            else:
                log.warn('Service %r has no "__packages__" defined', cls.__key__)

    @property
    def formats(self):
        return self._formats

    @property
    def plugins(self):
        return self._plugins

    @property
    def provider(self):
        return self._provider

    def load_all(self):
        for service in six.itervalues(self._services):
            service.load()

    def database_name(self, source, target):
        return self._databases.get((source, target))

    def package_name(self, source, target):
        return self._packages.get((source, target))

    def __getitem__(self, source):
        return ServiceInterface(self, source)

    #
    # Private methods
    #

    def _construct_services(self, services):
        result = {}

        for _, cls in self._load_plugins('client', services, construct=False):
            # Add supported service conversions
            for source, targets in cls.__services__.items():
                for target in targets:
                    # Construct service
                    result[(source, target)] = cls(self, source, target)

        return result

    def _construct_provider(self, provider_or_key):
        if isinstance(provider_or_key, Provider):
            # Class
            provider = provider_or_key
        elif isinstance(provider_or_key, six.string_types):
            # Identifier
            provider = PluginManager.get('client-provider', provider_or_key)

            if provider is None:
                raise ValueError('Unable to find provider: %r' % provider_or_key)
        else:
            raise ValueError('Unknown provider: %r' % provider_or_key)

        # Ensure provider has been constructed
        if inspect.isclass(provider):
            provider = provider()

        # Initialize provider
        provider.initialize(self)
        return provider

    @staticmethod
    def _load_plugins(kind, keys, construct=True):
        if not keys:
            return

        for name in keys:
            cls = PluginManager.get(kind, name)

            if cls is None:
                log.warn('Unable to find plugin: %r', name)
                continue

            if not cls.available:
                log.warn('Plugin %r is not available', name)
                continue

            if construct:
                yield cls.__key__, cls()
            else:
                yield cls.__key__, cls


class ServiceInterface(object):
    def __init__(self, client, source):
        self.client = client
        self.source = source

    def to(self, target):
        try:
            return self.client._services[(self.source, target)]
        except KeyError:
            raise KeyError('Unknown service: %s -> %s' % (self.source, target))



from oem.media.core.base.identifier import Identifier


class MovieIdentifier(Identifier):
    def __init__(self, part=None, progress=None):
        self.part = part

        self.progress = progress

    @property
    def valid(self):
        return True

    def to_dict(self):
        result = {}

        if self.progress is not None:
            result['progress'] = self.progress

        return result

    def __repr__(self):
        attributes = [
            ('%s: %r' % (key, getattr(self, key))) for key in ['progress'] if getattr(self, key)
        ]

        fragments = []

        if attributes:
            fragments.append('(%s)' % (', '.join(attributes)))

        return '<%s%s>' % (
            self.__class__.__name__,
            (' %s' % (' '.join(fragments))) if fragments else ''
        )
